from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from monarch_py.datamodels.solr import HistoPhenoKeys, SolrQuery
from monarch_py.datamodels.category_enums import AssociationPredicate
from monarch_py.utils.association_type_utils import AssociationTypeMappings, get_solr_query_fragment
from monarch_py.utils.utils import escape


@dataclass
class AssociationCountSuffixes:
    """Suffix strings used to build and parse facet queries for association counts."""

    direct_subject: str
    direct_object: str
    closure_subject: str
    closure_object: str
    ortho_subject: Optional[str] = None
    ortho_object: Optional[str] = None

    @property
    def all_suffixes(self) -> Dict[str, str]:
        """Return a mapping of count_level+direction to suffix string."""
        result = {
            "direct_subject": self.direct_subject,
            "direct_object": self.direct_object,
            "closure_subject": self.closure_subject,
            "closure_object": self.closure_object,
        }
        if self.ortho_subject:
            result["orthologs_subject"] = self.ortho_subject
        if self.ortho_object:
            result["orthologs_object"] = self.ortho_object
        return result


def build_association_count_suffixes(entities: List[str]) -> AssociationCountSuffixes:
    """Build the suffix strings used in facet queries for association counts.

    Shared between query building and response parsing to keep them in sync.

    Args:
        entities: List of entity IDs (primary entity first, then orthologs).
    """
    entity = entities[0]

    direct_subject = f'AND subject:"{entity}"'
    direct_object = f'AND (object:"{entity}" OR disease_context_qualifier:"{entity}")'
    closure_subject = f'AND (subject:"{entity}" OR subject_closure:"{entity}")'
    closure_object = f'AND (object:"{entity}" OR object_closure:"{entity}" OR disease_context_qualifier:"{entity}" OR disease_context_qualifier_closure:"{entity}")'

    ortho_subject = None
    ortho_object = None
    if len(entities) > 1:
        all_subjects = " OR ".join(f'subject:"{e}" OR subject_closure:"{e}"' for e in entities)
        all_objects = " OR ".join(
            f'object:"{e}" OR object_closure:"{e}" OR disease_context_qualifier:"{e}" OR disease_context_qualifier_closure:"{e}"'
            for e in entities
        )
        ortho_subject = f"AND ({all_subjects})"
        ortho_object = f"AND ({all_objects})"

    return AssociationCountSuffixes(
        direct_subject=direct_subject,
        direct_object=direct_object,
        closure_subject=closure_subject,
        closure_object=closure_object,
        ortho_subject=ortho_subject,
        ortho_object=ortho_object,
    )


def build_association_query(
    category: Optional[List[str]] = None,
    subject: Optional[List[str]] = None,
    subject_closure: Optional[str] = None,
    subject_category: Optional[List[str]] = None,
    subject_namespace: Optional[List[str]] = None,
    subject_taxon: Optional[List[str]] = None,
    predicate: Optional[List[str]] = None,
    object: Optional[List[str]] = None,
    object_closure: Optional[str] = None,
    object_category: Optional[List[str]] = None,
    object_namespace: Optional[List[str]] = None,
    object_taxon: Optional[List[str]] = None,
    entity: Optional[List[str]] = None,
    primary_knowledge_source: Optional[List[str]] = None,
    direct: bool = False,
    q: Optional[str] = None,
    sort: Optional[List[str]] = None,
    facet_fields: Optional[List[str]] = None,
    facet_queries: Optional[List[str]] = None,
    filter_queries: Optional[List[str]] = None,
    facet_mincount: int = 1,
    offset: int = 0,
    limit: int = 20,
) -> SolrQuery:
    """Populate a SolrQuery object with association filters"""
    entity_fields = ["subject", "object", "disease_context_qualifier"]
    query = SolrQuery(start=offset, rows=limit)
    query.add_field_filter_query("category", None if not category else [c for c in category])
    query.add_field_filter_query("predicate", None if not predicate else [p for p in predicate])
    query.add_field_filter_query("subject_closure", subject_closure)
    query.add_field_filter_query("subject_category", None if not subject_category else [c for c in subject_category])
    query.add_field_filter_query("subject_namespace", subject_namespace)
    query.add_field_filter_query("subject_taxon", subject_taxon)
    query.add_field_filter_query("object_closure", object_closure)
    query.add_field_filter_query("object_category", None if not object_category else [c for c in object_category])
    query.add_field_filter_query("object_namespace", object_namespace)
    query.add_field_filter_query("object_taxon", object_taxon)
    query.add_field_filter_query("primary_knowledge_source", primary_knowledge_source)
    if subject:
        if direct:
            query.add_field_filter_query("subject", " OR ".join(subject))
        else:
            query.add_filter_query(" OR ".join([f'subject:"{s}" OR subject_closure:"{s}"' for s in subject]))
    if object:
        if direct:
            query.add_field_filter_query("object", " OR ".join(object))
        else:
            query.add_filter_query(" OR ".join([f'object:"{o}" OR object_closure:"{o}"' for o in object]))
    if entity:
        if direct:
            query.add_filter_query(" OR ".join([f'{field}:"{e}"' for e in entity for field in entity_fields]))
        else:
            query.add_filter_query(
                " OR ".join([f'{field}:"{e}" OR {field}_closure:"{e}"' for e in entity for field in entity_fields])
            )
    if q:
        # We don't yet have tokenization strategies for the association index, initially we'll limit searching to
        # the visible fields in an association table plus their ID equivalents and use a wildcard query for substring matching
        query.q = q
        query.def_type = "edismax"
        query.hl = True
        query.query_fields = association_search_query_fields()
    if sort:
        query.sort = ", ".join(sort)
    if facet_fields:
        query.facet_fields = facet_fields
    if facet_queries:
        query.facet_queries = facet_queries
    if filter_queries:
        query.filter_queries.extend(filter_queries)
    query.facet_mincount = facet_mincount
    return query


def build_association_table_query(
    entity: List[str],
    category: str,
    direct: bool = False,
    q: Optional[str] = None,
    facet_fields: List[str] = None,
    facet_queries: List[str] = None,
    filter_queries: List[str] = None,
    offset: int = 0,
    limit: int = 5,
    sort: List[str] = None,
) -> SolrQuery:
    if sort is None:
        sort = [
            "frequency_computed_sortable_float desc",
            "evidence_count desc",
            "subject_label asc",
            "predicate asc",
            "object_label asc",
            "primary_knowledge_source asc",
        ]

    query = build_association_query(
        entity=entity,
        category=[category],
        q=q,
        sort=sort,
        offset=offset,
        limit=limit,
        direct=direct,
        facet_fields=facet_fields,
        facet_queries=facet_queries,
        filter_queries=filter_queries,
    )
    return query


def build_association_counts_query(entities: List[str]) -> SolrQuery:
    """Build a query that produces three levels of association counts via facet queries.

    Args:
        entities: List of entity IDs. The first is the primary entity; the rest are orthologs.
                  If only one entity is provided, ortholog counts are skipped.
    """
    suffixes = build_association_count_suffixes(entities)

    facet_queries = []
    for suffix in suffixes.all_suffixes.values():
        for agm in AssociationTypeMappings.get_mappings():
            association_type_query = get_solr_query_fragment(agm)
            facet_queries.append(f"({association_type_query}) {suffix}")

    query = build_association_query(entity=entities, facet_queries=facet_queries)
    return query


def build_histopheno_query(subject: str) -> SolrQuery:
    query = build_association_query(
        subject=[subject],
        offset=0,
        limit=0,
    )
    hpkeys = [i for i in HistoPhenoKeys]
    query.facet_queries = [f'object_closure:"{(i.value)}"' for i in hpkeys]
    return query


def build_multi_entity_association_query(
    entity: str,
    counterpart_category: Optional[str] = None,
    # predicate: List[str] = None,
    offset: int = 0,
    limit: int = 20,
) -> SolrQuery:
    """Populate a SolrQuery object with association filters"""
    query = SolrQuery(start=offset, rows=limit)
    if counterpart_category:
        query.add_filter_query(
            f'(subject:"{escape(entity)}" AND object_category:"{escape(counterpart_category)}") OR (object:"{escape(entity)}" AND subject_category:"{escape(counterpart_category)}")'
        )
    else:
        query.add_filter_query(f'(subject:"{escape(entity)}") OR (object:"{escape(entity)}")')
    return query


def build_search_query(
    q: str = "*:*",
    offset: int = 0,
    limit: int = 20,
    category: List[str] = None,
    in_taxon_label: List[str] = None,
    facet_fields: List[str] = None,
    facet_queries: List[str] = None,
    filter_queries: List[str] = None,
    highlighting: bool = False,
    sort: Optional[str] = None,
) -> SolrQuery:
    query = SolrQuery(start=offset, rows=limit, sort=sort)
    query.q = q
    query.def_type = "edismax"
    query.query_fields = entity_query_fields()
    query.hl = highlighting
    query.boost = entity_boost(empty_search=(q == "*:*"))
    if category:
        query.add_filter_query(" OR ".join(f'category:"{cat}"' for cat in category))
    if in_taxon_label:
        query.add_filter_query(" OR ".join([f'in_taxon_label:"{t}"' for t in in_taxon_label]))
    if facet_fields:
        query.facet_fields = facet_fields
    if facet_queries:
        query.facet_queries = facet_queries
    if filter_queries:
        query.filter_queries.extend(filter_queries)
    # Filter out entities that don't have names (required field, but some don't have them)
    query.add_filter_query("name:*")
    return query


def build_autocomplete_query(
    q: str, category: List[str] = None, prioritized_predicates: List[AssociationPredicate] = None
) -> SolrQuery:
    query = SolrQuery(q=q, rows=10, start=0)
    query.q = q
    if category:
        query.add_filter_query(" OR ".join(f'category:"{cat}"' for cat in category))
    # match the query fields to start with
    query.query_fields = entity_query_fields()
    query.def_type = "edismax"
    query.boost = entity_boost(prioritized_predicates=prioritized_predicates, empty_search=(q == "*:*"))
    return query


def build_mapping_query(
    entity_id: Optional[List[str]] = None,
    subject_id: Optional[List[str]] = None,
    predicate_id: Optional[List[str]] = None,
    object_id: Optional[List[str]] = None,
    mapping_justification: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 20,
) -> SolrQuery:
    query = SolrQuery(start=offset, rows=limit)
    if entity_id:
        query.add_filter_query(" OR ".join([f'subject_id:"{escape(i)}" OR object_id:"{escape(i)}"' for i in entity_id]))
    if subject_id:
        query.add_filter_query(" OR ".join([f'subject_id:"{escape(i)}"' for i in subject_id]))
    if predicate_id:
        query.add_filter_query(" OR ".join([f'predicate_id:"{escape(i)}"' for i in predicate_id]))
    if object_id:
        query.add_filter_query(" OR ".join([f'object_id:"{escape(i)}"' for i in object_id]))
    if mapping_justification:
        query.add_filter_query(" OR ".join([f'mapping_justification:"{escape(i)}"' for i in mapping_justification]))
    return query


def build_grounding_query(text: str) -> SolrQuery:
    query = SolrQuery(q=text, rows=10, start=0)
    query.q = f'"{text}"'  # quoting so that the complete text is matched as a unit
    # rather than _t (text) or _ac (autocomplete/starts-with), just use keyword fields
    query.query_fields = (
        "id^100 name^10 symbol^10 synonym name_grounding full_name_grounding symbol_grounding synonym_grounding"
    )
    query.def_type = "edismax"
    query.boost = obsolete_unboost(multiplier=0.001)
    return query


### Search helper functions ###


def obsolete_unboost(multiplier=0.1):
    return f'if(termfreq(deprecated,"true"),{multiplier},1)'


def entity_boost(prioritized_predicates: List[AssociationPredicate] = None, empty_search: bool = False) -> str:
    """Shared boost function between search and autocomplete"""
    phenotype_boost = category_boost("biolink:PhenotypicFeature", 1.1)
    disease_boost = category_boost("biolink:Disease", 1.3)
    human_gene_boost = category_boost("biolink:Gene", 1.1, taxon="NCBITaxon:9606")

    boosts = [phenotype_boost, disease_boost, human_gene_boost, obsolete_unboost()]
    if prioritized_predicates:
        boosts.append(entity_predicate_boost(prioritized_predicates, 2.0))
    if empty_search:
        boosts.append(blank_search_boost())
    return f"product({','.join(boosts)})"


def entity_predicate_boost(prioritized_predicates: List[AssociationPredicate], multiplier: float) -> str:
    boosts = []
    for predicate in prioritized_predicates:
        field_root = predicate.value.replace("biolink:", "")
        count_field = field_root + "_count"
        boosts.append(f"if({count_field},{multiplier},1)")
    return ",".join(boosts)


def category_boost(category: str, multiplier: float, taxon: Optional[str] = None) -> str:
    if taxon:
        return f'if(and(termfreq(in_taxon,"{taxon}"),termfreq(category,"{category}")),{multiplier},1)'
    else:
        return f'if(termfreq(category,"{category}"),{multiplier},1)'


def blank_search_boost() -> str:
    """
    Boost specific nodes that we'd like to see as site examples to the top for empty searches
    """
    example_nodes = [
        "MONDO:0007523",  # Ehlers-Danlos syndrome, hypermobility type
        "MONDO:0019391",  # Fanconi anemia
        "MONDO:0018954",  # Loeys-Dietz syndrome
        "MONDO:0011518",  # Wiedemann-Steiner syndrome
        "HP:0001166",  # Arachnodactyly
        "HP:0001631",  # Atrial septal defect
        "UBERON:0000948",  # heart
        "UBERON:0006585",  # vestibular organ
        "HGNC:4851",  # HTT
        "HGNC:3603",  # FBN1
    ]
    # boost score by 1.5 + i/10 for these nodes
    boosts = [f'if(termfreq(id,"{node}"),{len(example_nodes) - i + 2},1)' for i, node in enumerate(example_nodes)]
    boost = ",".join(boosts)
    return f"product({boost})"


def autocomplete_query_fields():
    """
    Fields and boosts used for autocomplete
    """
    return "id^100 name^10 name_t^5 name_ac symbol^10 symbol_t^5 symbol_ac synonym synonym_t synonym_ac in_taxon_label_t"


def entity_query_fields():
    """
    Fields and boosts used for entity search, includes autocomplete fields and expands upon them beyond pure name fields
    """
    return f"{autocomplete_query_fields()} description_t xref"


def association_search_query_fields():
    """
    Shared field list for free text search on associations (e.g. for the association table)
    """

    return (
        "subject subject_label^2 subject_label_t subject_closure subject_closure_label subject_closure_label_t"
        " predicate predicate_t"
        " object object_label^2 object_label_t object_closure object_closure_label object_closure_label_t"
        " publications has_evidence primary_knowledge_source aggregator_knowledge_source provided_by "
    )


def build_case_phenotype_query(
    disease_id: str,
    direct_only: bool,
    rows: int = 50000,
) -> Dict[str, Any]:
    """Build Solr query for case-phenotype matrix.

    Uses Solr's JOIN query parser to find all phenotype associations
    for cases that have a specific disease association.

    The query structure:
    1. Main query: JOIN from CaseToDiseaseAssociation to CaseToPhenotypicFeatureAssociation
       - Joins on `subject` field (the case ID)
       - Inner query filters diseases by object (direct) or object_closure (indirect)
    2. Filter query: Restrict results to phenotype associations only
    3. Facet queries: Count phenotypes per HistoPheno bin

    Args:
        disease_id: MONDO disease ID to query
        direct_only: If True, only cases with exactly this disease;
                     If False, includes cases with descendant diseases
        rows: Maximum results to return

    Returns:
        Dict of Solr query parameters ready for requests

    Example direct query:
        {!join from=subject to=subject}(
            category:"biolink:CaseToDiseaseAssociation"
            AND object:"MONDO:0007078"
        )

    Example indirect query (via closure):
        {!join from=subject to=subject}(
            category:"biolink:CaseToDiseaseAssociation"
            AND object_closure:"MONDO:0007078"
        )
    """
    disease_field = "object" if direct_only else "object_closure"

    join_query = (
        '{!join from=subject to=subject}'
        f'(category:"biolink:CaseToDiseaseAssociation" '
        f'AND {disease_field}:"{escape(disease_id)}")'
    )

    facet_queries = [
        f'object_closure:"{bin_key.value}"'
        for bin_key in HistoPhenoKeys
    ]

    return {
        "q": join_query,
        "fq": 'category:"biolink:CaseToPhenotypicFeatureAssociation"',
        "rows": rows,
        "fl": ",".join([
            "subject",           # Case ID
            "subject_label",     # Case label
            "object",            # Phenotype ID
            "object_label",      # Phenotype label
            "object_closure",    # For bin assignment
            "negated",           # Explicit absence
            "publications",      # Supporting evidence
            "onset_qualifier",   # Age of onset (ISO8601)
            "onset_qualifier_label",
        ]),
        "facet": "true",
        "facet.query": facet_queries,
    }


def build_case_disease_query(
    disease_id: str,
    direct_only: bool,
    rows: int = 50000,
) -> Dict[str, Any]:
    """Build Solr query to get cases for a disease.

    This is a simpler query to get CaseToDiseaseAssociation documents,
    which we need to determine:
    1. Total case count (for limit checking before full matrix fetch)
    2. Which specific disease each case has (for direct/indirect flagging)

    Args:
        disease_id: MONDO disease ID to query
        direct_only: If True, only cases with exactly this disease
        rows: Maximum results to return

    Returns:
        Dict of Solr query parameters
    """
    disease_field = "object" if direct_only else "object_closure"

    return {
        "q": f'{disease_field}:"{escape(disease_id)}"',
        "fq": 'category:"biolink:CaseToDiseaseAssociation"',
        "rows": rows,
        "fl": "subject,subject_label,object,object_label",
    }


def build_grid_column_query(
    context_id: str,
    config: "GridTypeConfig",
    direct_only: bool,
    filter_empty_columns: bool = True,
    rows: int = 50000,
) -> Dict[str, Any]:
    """Build Solr query to get column entities for a grid.

    First hop: Context Entity -> Column Entities

    Args:
        context_id: The context entity ID (e.g., disease ID, gene ID)
        config: Grid type configuration
        direct_only: If True, only direct associations; if False, use closure
        filter_empty_columns: If True, exclude columns with no row associations
        rows: Maximum results to return

    Returns:
        Dict of Solr query parameters
    """
    from monarch_py.datamodels.grid_configs import GridTypeConfig

    context_field = config.context_field if direct_only else config.context_closure_field

    # Base filter for column association category (handles single or multi-category configs)
    column_categories = config.get_column_assoc_categories()
    if len(column_categories) == 1:
        category_filter = f'category:"{column_categories[0].value}"'
    else:
        category_filter = "(" + " OR ".join(f'category:"{c.value}"' for c in column_categories) + ")"
    fq = [category_filter]

    # Build main query
    if filter_empty_columns:
        # Use JOIN to find column associations where the column entity
        # appears in row associations. This filters out columns with no rows.
        # JOIN: from=row_context_field (e.g., subject in phenotype assocs)
        #       to=column_field (e.g., object in homology assocs)
        existence_join = (
            f'{{!join from={config.row_context_field} to={config.column_field}}}'
            f'category:"{config.row_assoc_category.value}"'
        )
        q = existence_join
        # Add context filter to fq instead of q
        fq.append(f'{context_field}:"{context_id}"')
    else:
        q = f'{context_field}:"{context_id}"'

    return {
        "q": q,
        "fq": fq,
        "rows": rows,
        "fl": ",".join([
            config.column_field,
            f"{config.column_field}_label",
            config.context_field,
            f"{config.context_field}_label",
            "subject_taxon",
            "subject_taxon_label",
            "object_taxon",
            "object_taxon_label",
        ]),
    }


def build_multi_category_column_query(
    context_id: str,
    column_assoc_categories: List[str],
    context_field: str,
    context_closure_field: str,
    column_field: str,
    direct_only: bool,
    row_assoc_categories: Optional[List[str]] = None,
    row_context_field: Optional[str] = None,
    filter_empty_columns: bool = True,
    rows: int = 50000,
) -> Dict[str, Any]:
    """Build Solr query to get column entities from multiple association categories.

    First hop: Context Entity -> Column Entities (with multiple category support)

    Args:
        context_id: The context entity ID (e.g., gene ID)
        column_assoc_categories: List of association category values
        context_field: Field in associations containing context ID
        context_closure_field: Field for indirect context matching
        column_field: Field containing column entity ID
        direct_only: If True, only direct associations; if False, use closure
        row_assoc_categories: List of row association categories (for empty column filtering)
        row_context_field: Field in row associations linking to columns (for filtering)
        filter_empty_columns: If True, exclude columns with no row associations
        rows: Maximum results to return

    Returns:
        Dict of Solr query parameters
    """
    ctx_field = context_field if direct_only else context_closure_field

    # Build OR query for multiple column categories
    column_category_filter = " OR ".join(f'category:"{cat}"' for cat in column_assoc_categories)

    # Base filter queries
    fq = [f'({column_category_filter})']

    # Build main query
    if filter_empty_columns and row_assoc_categories and row_context_field:
        # Use JOIN to find column associations where the column entity
        # appears in row associations. This filters out columns with no rows.
        row_category_filter = " OR ".join(f'category:"{cat}"' for cat in row_assoc_categories)
        existence_join = (
            f'{{!join from={row_context_field} to={column_field}}}'
            f'({row_category_filter})'
        )
        q = existence_join
        # Add context filter to fq instead of q
        fq.append(f'{ctx_field}:"{context_id}"')
    else:
        q = f'{ctx_field}:"{context_id}"'

    return {
        "q": q,
        "fq": fq,
        "rows": rows,
        "fl": ",".join([
            column_field,
            f"{column_field}_label",
            context_field,
            f"{context_field}_label",
            "subject_taxon",
            "subject_taxon_label",
            "object_taxon",
            "object_taxon_label",
            # Include source association fields
            "category",
            "predicate",
            "publications",
            "has_evidence",
            "primary_knowledge_source",
        ]),
    }


def build_multi_category_row_query(
    context_id: str,
    column_assoc_categories: List[str],
    row_assoc_categories: List[str],
    context_field: str,
    context_closure_field: str,
    column_field: str,
    row_context_field: str,
    row_entity_field: str,
    grouping: "RowGroupingConfig",
    direct_only: bool,
    rows: int = 50000,
) -> Dict[str, Any]:
    """Build Solr query for grid row entities using JOIN with multiple column categories.

    Second hop: Column Entities -> Row Entities (supporting multiple column categories)

    Args:
        context_id: The context entity ID
        column_assoc_categories: List of association categories for the column hop
        row_assoc_categories: List of association categories for the row hop
        context_field: Field in column associations containing context ID
        context_closure_field: Field for indirect context matching
        column_field: Field containing column entity ID
        row_context_field: Field in row associations to join on
        row_entity_field: Field containing row entity ID
        grouping: Row grouping configuration (for facet queries)
        direct_only: If True, only direct associations; if False, use closure
        rows: Maximum results to return

    Returns:
        Dict of Solr query parameters
    """
    from monarch_py.datamodels.grid_groupings import RowGroupingConfig

    ctx_field = context_field if direct_only else context_closure_field

    # Build OR query for multiple categories in the JOIN
    category_filter = " OR ".join(f'category:"{cat}"' for cat in column_assoc_categories)

    # JOIN query: Find row associations for column entities associated with context
    join_query = (
        f'{{!join from={column_field} to={row_context_field}}}'
        f'(({category_filter}) '
        f'AND {ctx_field}:"{context_id}")'
    )

    # Build facet queries for bin counts
    facet_queries = [
        f'{row_entity_field}_closure:"{bin_id}"'
        for bin_id in grouping.bin_ids
    ]

    # Build OR query for multiple row categories
    row_category_filter = " OR ".join(f'category:"{cat}"' for cat in row_assoc_categories)

    return {
        "q": join_query,
        "fq": f'({row_category_filter})',
        "rows": rows,
        "fl": ",".join([
            row_context_field,                      # Column entity ID
            f"{row_context_field}_label",           # Column entity label
            row_entity_field,                       # Row entity ID
            f"{row_entity_field}_label",            # Row entity label
            f"{row_entity_field}_closure",          # For bin assignment
            "negated",
            "publications",
            "onset_qualifier",
            "onset_qualifier_label",
        ]),
        "facet": "true",
        "facet.query": facet_queries,
    }


def build_grid_row_query(
    context_id: str,
    config: "GridTypeConfig",
    grouping: "RowGroupingConfig",
    direct_only: bool,
    rows: int = 50000,
) -> Dict[str, Any]:
    """Build Solr query for grid row entities using JOIN.

    Second hop: Column Entities -> Row Entities

    Uses Solr's JOIN query parser to find all row associations
    for column entities that are associated with the context entity.

    Args:
        context_id: The context entity ID
        config: Grid type configuration
        grouping: Row grouping configuration (for facet queries)
        direct_only: If True, only direct associations; if False, use closure
        rows: Maximum results to return

    Returns:
        Dict of Solr query parameters
    """
    from monarch_py.datamodels.grid_configs import GridTypeConfig
    from monarch_py.datamodels.grid_groupings import RowGroupingConfig

    context_field = config.context_field if direct_only else config.context_closure_field

    # Build column category filter (handles single or multi-category configs)
    column_categories = config.get_column_assoc_categories()
    if len(column_categories) == 1:
        col_cat_filter = f'category:"{column_categories[0].value}"'
    else:
        col_cat_filter = "(" + " OR ".join(f'category:"{c.value}"' for c in column_categories) + ")"

    # JOIN query: Find row associations for column entities associated with context
    join_query = (
        f'{{!join from={config.column_field} to={config.row_context_field}}}'
        f'({col_cat_filter} '
        f'AND {context_field}:"{context_id}")'
    )

    # Build facet queries for bin counts
    facet_queries = [
        f'{config.row_entity_field}_closure:"{bin_id}"'
        for bin_id in grouping.bin_ids
    ]

    return {
        "q": join_query,
        "fq": f'category:"{config.row_assoc_category.value}"',
        "rows": rows,
        "fl": ",".join([
            config.row_context_field,                      # Column entity ID
            f"{config.row_context_field}_label",           # Column entity label
            config.row_entity_field,                       # Row entity ID
            f"{config.row_entity_field}_label",            # Row entity label
            f"{config.row_entity_field}_closure",          # For bin assignment
            "negated",
            "publications",
            "onset_qualifier",
            "onset_qualifier_label",
        ]),
        "facet": "true",
        "facet.query": facet_queries,
    }



"""Unit tests for case_phenotype API endpoints."""

import pytest
from unittest.mock import patch, MagicMock
from fastapi import HTTPException
from fastapi.testclient import TestClient

from monarch_py.api.case_phenotype import validate_disease_id, _is_disease_entity
from monarch_py.api.main import app


# =====================================================================
# Tests for validate_disease_id
# =====================================================================


@pytest.mark.parametrize(
    "input_id",
    ["MONDO:0007078", "mondo:0007078", "MONDO_0007078", "  MONDO:0007078  "],
    ids=["canonical", "lowercase", "underscore", "whitespace"],
)
def test_validate_disease_id_normalizes(input_id):
    assert validate_disease_id(input_id) == "MONDO:0007078"


@pytest.mark.parametrize(
    "input_id",
    ["", "HP:0001234", "MONDO:123", "MONDO:12345678"],
    ids=["empty", "wrong-prefix", "too-short", "too-long"],
)
def test_validate_disease_id_rejects_invalid(input_id):
    with pytest.raises(HTTPException) as exc_info:
        validate_disease_id(input_id)
    assert exc_info.value.status_code == 422


# =====================================================================
# Tests for _is_disease_entity
# =====================================================================


def _make_entity_with_category(category):
    """Helper to create a MagicMock entity with a given category."""
    if category is _SENTINEL_NO_ATTR:
        return MagicMock(spec=[])  # No category attribute
    entity = MagicMock()
    entity.category = category
    return entity


_SENTINEL_NO_ATTR = object()


@pytest.mark.parametrize(
    "category,expected",
    [
        ("biolink:Disease", True),
        (["biolink:Disease", "biolink:NamedThing"], True),
        ("biolink:DiseaseOrPhenotypicFeature", True),
        ("biolink:Gene", False),
        (_SENTINEL_NO_ATTR, False),
        (None, False),
    ],
    ids=["string", "list", "disease-or-phenotypic", "non-disease", "no-attr", "none"],
)
def test_is_disease_entity(category, expected):
    entity = _make_entity_with_category(category)
    assert _is_disease_entity(entity) is expected


# =====================================================================
# Tests for GET /case-phenotype-matrix/{disease_id} endpoint
# =====================================================================


@pytest.fixture
def client():
    return TestClient(app)


def test_case_phenotype_matrix_valid_disease(client):
    from monarch_py.implementations.solr.solr_implementation import SolrImplementation
    from monarch_py.datamodels.model import CasePhenotypeMatrixResponse

    mock_entity = MagicMock()
    mock_entity.category = "biolink:Disease"

    mock_response = CasePhenotypeMatrixResponse(
        disease_id="MONDO:0007078",
        disease_name="Achondroplasia",
        total_cases=2,
        total_phenotypes=5,
        cases=[],
        phenotypes=[],
        bins=[],
        cells={},
    )

    with (
        patch.object(SolrImplementation, "get_entity", return_value=mock_entity),
        patch.object(SolrImplementation, "get_case_phenotype_matrix", return_value=mock_response),
    ):
        response = client.get("/v3/api/case-phenotype-matrix/MONDO:0007078")
        assert response.status_code == 200
        data = response.json()
        assert data["disease_id"] == "MONDO:0007078"


def test_case_phenotype_matrix_invalid_id(client):
    response = client.get("/v3/api/case-phenotype-matrix/INVALID")
    assert response.status_code == 422


def test_case_phenotype_matrix_not_found(client):
    from monarch_py.implementations.solr.solr_implementation import SolrImplementation

    with patch.object(SolrImplementation, "get_entity", return_value=None):
        response = client.get("/v3/api/case-phenotype-matrix/MONDO:9999999")
        assert response.status_code == 404


def test_case_phenotype_matrix_non_disease(client):
    from monarch_py.implementations.solr.solr_implementation import SolrImplementation

    mock_entity = MagicMock()
    mock_entity.category = "biolink:Gene"

    with patch.object(SolrImplementation, "get_entity", return_value=mock_entity):
        response = client.get("/v3/api/case-phenotype-matrix/MONDO:0007078")
        assert response.status_code == 400
        assert "not a disease" in response.json()["detail"].lower()


def test_case_phenotype_matrix_limit_exceeded(client):
    from monarch_py.implementations.solr.solr_implementation import SolrImplementation

    mock_entity = MagicMock()
    mock_entity.category = "biolink:Disease"

    with (
        patch.object(SolrImplementation, "get_entity", return_value=mock_entity),
        patch.object(SolrImplementation, "get_case_phenotype_matrix", side_effect=ValueError("exceeds limit")),
    ):
        response = client.get("/v3/api/case-phenotype-matrix/MONDO:0007078")
        assert response.status_code == 400



"""Unit tests for grid data models."""
import pytest
from monarch_py.datamodels.model import GridColumnEntity


class TestGridColumnEntity:
    """Test GridColumnEntity model with source association fields."""

    def test_source_association_fields_optional(self):
        """Source association fields should be optional."""
        entity = GridColumnEntity(
            id="MONDO:0007078",
            category="biolink:Disease",
            is_direct=True,
        )
        assert entity.id == "MONDO:0007078"
        assert entity.source_association_category is None
        assert entity.source_association_predicate is None
        assert entity.source_association_publications is None
        assert entity.source_association_evidence_count is None

    def test_source_association_fields_populated(self):
        """Source association fields can be populated."""
        entity = GridColumnEntity(
            id="MONDO:0007078",
            label="Achondroplasia",
            category="biolink:Disease",
            is_direct=True,
            source_association_category="biolink:CausalGeneToDiseaseAssociation",
            source_association_predicate="biolink:causes",
            source_association_publications=["PMID:12345", "PMID:67890"],
            source_association_evidence_count=5,
            source_association_primary_knowledge_source="infores:orphanet",
        )
        assert entity.source_association_category == "biolink:CausalGeneToDiseaseAssociation"
        assert entity.source_association_predicate == "biolink:causes"
        assert entity.source_association_publications == ["PMID:12345", "PMID:67890"]
        assert entity.source_association_evidence_count == 5
        assert entity.source_association_primary_knowledge_source == "infores:orphanet"

    def test_json_serialization_includes_source_association(self):
        """Model should serialize source association fields to JSON."""
        entity = GridColumnEntity(
            id="MONDO:0007078",
            category="biolink:Disease",
            is_direct=True,
            source_association_category="biolink:CausalGeneToDiseaseAssociation",
        )
        data = entity.model_dump()
        assert "source_association_category" in data
        assert data["source_association_category"] == "biolink:CausalGeneToDiseaseAssociation"



"""API endpoint for case-phenotype matrix."""
import logging
import re
from fastapi import APIRouter, HTTPException, Query, Path

from monarch_py.api.config import solr
from monarch_py.datamodels.model import CasePhenotypeMatrixResponse, Node

logger = logging.getLogger(__name__)

router = APIRouter(
    tags=["case-phenotype"],
    responses={
        400: {"description": "Case count exceeds limit, or ID is not a disease"},
        404: {"description": "Disease not found"},
        422: {"description": "Invalid disease ID format"},
    },
)

DISEASE_ID_PATTERN = re.compile(r"^MONDO:\d{7}$")


def validate_disease_id(disease_id: str) -> str:
    """Validate and normalize disease ID.

    Args:
        disease_id: The disease ID to validate

    Returns:
        Normalized disease ID

    Raises:
        HTTPException: If ID is invalid
    """
    if not disease_id:
        raise HTTPException(
            status_code=422,
            detail="Disease ID is required",
        )

    normalized = disease_id.strip().upper()
    if "_" in normalized and ":" not in normalized:
        normalized = normalized.replace("_", ":", 1)

    if not DISEASE_ID_PATTERN.match(normalized):
        raise HTTPException(
            status_code=422,
            detail=f"Invalid disease ID format: '{disease_id}'. Expected format: MONDO:0000000",
        )

    return normalized


def _is_disease_entity(entity: Node) -> bool:
    """Check if entity is a disease.

    Args:
        entity: Entity object to check

    Returns:
        True if entity is a disease
    """
    category = getattr(entity, "category", None)
    if category is None:
        return False

    # Category can be a string or list
    categories = [category] if isinstance(category, str) else category

    disease_categories = {"biolink:Disease", "biolink:DiseaseOrPhenotypicFeature"}
    return bool(set(categories) & disease_categories)


@router.get(
    "/{disease_id}",
    response_model=CasePhenotypeMatrixResponse,
    summary="Get case-phenotype matrix for a disease",
    description="""
    Returns a matrix of cases (patients) and their phenotypes for a given disease.

    The matrix includes:
    - **cases**: List of patient cases with the disease
    - **phenotypes**: List of phenotypes observed across all cases
    - **bins**: Body system categories (HistoPheno) with phenotype counts
    - **cells**: Map of case:phenotype pairs to observation details

    Use `direct=true` (default) to get only cases with exactly this disease.
    Use `direct=false` to include cases with descendant diseases (subtypes).
    """,
)
async def get_case_phenotype_matrix(
    disease_id: str = Path(
        ...,
        description="MONDO disease ID",
        examples=["MONDO:0007078"],
    ),
    direct: bool = Query(
        True,
        description="Only include cases with exactly this disease",
    ),
    limit: int = Query(
        1000,
        ge=1,
        le=1000,
        description="Maximum number of cases",
    ),
) -> CasePhenotypeMatrixResponse:
    """Get case-phenotype matrix for a disease.

    Args:
        disease_id: MONDO disease ID
        direct: If True, only include cases with exactly this disease
        limit: Maximum number of cases allowed

    Returns:
        CasePhenotypeMatrixResponse with complete matrix data

    Raises:
        HTTPException: On validation errors or if limit exceeded
    """
    normalized_id = validate_disease_id(disease_id)

    solr_impl = solr()

    # Check if entity exists and is a disease
    try:
        entity = solr_impl.get_entity(normalized_id, extra=False)
        if entity is None:
            raise HTTPException(
                status_code=404,
                detail=f"Disease {normalized_id} not found",
            )

        if not _is_disease_entity(entity):
            raise HTTPException(
                status_code=400,
                detail=f"{normalized_id} is not a disease ID. Use a MONDO ID.",
            )
    except HTTPException:
        raise
    except Exception as e:
        logger.exception("Error looking up entity %s", normalized_id)
        raise HTTPException(
            status_code=500,
            detail=f"Error looking up disease {normalized_id}",
        ) from e

    # Get the matrix
    try:
        return solr_impl.get_case_phenotype_matrix(
            disease_id=normalized_id,
            direct_only=direct,
            limit=limit,
        )
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))



import pytest
from monarch_py.datamodels.model import AssociationTypeMapping
from monarch_py.utils.association_type_utils import (
    AssociationTypeMappings,
    get_association_type_mapping_by_query_string,
    get_solr_query_fragment,
    get_sql_query_fragment,
    parse_query_string_for_category,
)


@pytest.fixture()
def basic_mapping():
    return AssociationTypeMapping(
        subject_label="Genes",
        object_label="Phenotypes",
        category="biolink:GeneToPhenotypeAssociation",
    )


def test_solr_basic_mapping(basic_mapping):
    query_fragment = get_solr_query_fragment(basic_mapping)
    assert query_fragment == 'category:"biolink:GeneToPhenotypeAssociation"'


def test_sql_basic_mapping(basic_mapping):
    query_fragment = get_sql_query_fragment(basic_mapping)
    assert query_fragment == 'category = "biolink:GeneToPhenotypeAssociation"'


def test_parse_association_type_query_string_single_category():
    query_string = 'category:"biolink:GeneToPhenotypeAssociation"'
    category = parse_query_string_for_category(query_string)
    assert category == "biolink:GeneToPhenotypeAssociation"


def test_parse_association_type_query_string_multiple_categories():
    query_string = 'category:"biolink:GeneToDiseaseAssociation" AND (predicate:"biolink:gene_associated_with_condition" OR predicate:"biolink:contributes_to")'
    category = parse_query_string_for_category(query_string)
    assert category == "biolink:GeneToDiseaseAssociation"


# =====================================================================
# Tests for AssociationTypeMappings singleton
# =====================================================================


# =====================================================================
# Tests for AssociationTypeMappings singleton
# =====================================================================


def test_get_mappings_returns_list():
    mappings = AssociationTypeMappings.get_mappings()
    assert isinstance(mappings, list)
    assert len(mappings) > 0


def test_all_mappings_have_category():
    mappings = AssociationTypeMappings.get_mappings()
    for m in mappings:
        assert m.category is not None


def test_get_mapping_by_category():
    result = AssociationTypeMappings.get_mapping("biolink:DiseaseToPhenotypicFeatureAssociation")
    assert result is not None
    assert result.category == "biolink:DiseaseToPhenotypicFeatureAssociation"


def test_get_mapping_unknown_returns_none():
    result = AssociationTypeMappings.get_mapping("biolink:NonExistent")
    assert result is None


def test_get_traversable_associations_for_gene():
    results = AssociationTypeMappings.get_traversable_associations("biolink:Gene")
    assert len(results) > 0
    categories = [r["category"] for r in results]
    assert "biolink:GeneToPhenotypicFeatureAssociation" in categories


def test_get_traversable_associations_for_disease():
    results = AssociationTypeMappings.get_traversable_associations("biolink:Disease")
    assert len(results) > 0
    categories = [r["category"] for r in results]
    assert "biolink:DiseaseToPhenotypicFeatureAssociation" in categories


def test_get_traversable_associations_returns_direction():
    results = AssociationTypeMappings.get_traversable_associations("biolink:Gene")
    for r in results:
        assert "context_field" in r
        assert r["context_field"] in ("subject", "object")
        assert "target_category" in r
        assert "label" in r


def test_get_traversable_associations_empty_for_unknown():
    results = AssociationTypeMappings.get_traversable_associations("biolink:Unknown")
    assert results == []


# =====================================================================
# Tests for get_association_type_mapping_by_query_string
# =====================================================================


def test_mapping_by_query_string_valid():
    result = get_association_type_mapping_by_query_string('category:"biolink:DiseaseToPhenotypicFeatureAssociation"')
    assert result.category == "biolink:DiseaseToPhenotypicFeatureAssociation"


def test_mapping_by_query_string_no_match():
    with pytest.raises(ValueError, match="No matching"):
        get_association_type_mapping_by_query_string('category:"biolink:NonExistentAssociation"')


def test_mapping_by_query_string_no_category():
    with pytest.raises(ValueError, match="No categories"):
        get_association_type_mapping_by_query_string("predicate:something")



import pytest
from unittest.mock import patch, MagicMock

from monarch_py.implementations.solr.solr_implementation import SolrImplementation
from monarch_py.datamodels.model import EntityGridResponse, CasePhenotypeMatrixResponse


def test_get_counterpart_entities_limit():
    # Make sure that we don't accidentally end up with the default limit again
    with patch.object(SolrImplementation, "get_associations") as mock_get_associations:
        si = SolrImplementation()
        si.get_counterpart_entities("MONDO:0007947")
        assert mock_get_associations.called
        assert mock_get_associations.call_args_list[0].kwargs["limit"] == 1000


def test_get_generic_entity_grid_accepts_multiple_row_categories():
    """get_generic_entity_grid should accept row_assoc_categories as a list."""
    mock_entity = MagicMock()
    mock_entity.category = "biolink:Disease"
    mock_entity.configure_mock(name="Type 2 diabetes mellitus")

    with (
        patch.object(SolrImplementation, "get_entity", return_value=mock_entity),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Type 2 diabetes mellitus"),
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
    ):
        mock_query.return_value = {"response": {"docs": []}}

        si = SolrImplementation()

        # This should not raise TypeError - it should accept row_assoc_categories
        result = si.get_generic_entity_grid(
            context_id="MONDO:0005148",
            column_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
            row_assoc_categories=[
                "biolink:DiseaseToPhenotypicFeatureAssociation",
                "biolink:CaseToPhenotypicFeatureAssociation",
            ],
        )

        # Verify the call went through
        assert result is not None


# =====================================================================
# Tests for _get_entity_name
# =====================================================================


def _make_named_entity(name):
    entity = MagicMock()
    entity.configure_mock(name=name)
    return entity


def test_get_entity_name_returns_name():
    with patch.object(SolrImplementation, "get_entity", return_value=_make_named_entity("Huntington disease")):
        assert SolrImplementation()._get_entity_name("MONDO:0007078") == "Huntington disease"


@pytest.mark.parametrize(
    "get_entity_return,expected",
    [
        (None, "MONDO:0007078"),
        (_make_named_entity(None), "MONDO:0007078"),
    ],
    ids=["not-found", "name-is-none"],
)
def test_get_entity_name_fallback(get_entity_return, expected):
    with patch.object(SolrImplementation, "get_entity", return_value=get_entity_return):
        assert SolrImplementation()._get_entity_name("MONDO:0007078") == expected


def test_get_entity_name_on_exception():
    with patch.object(SolrImplementation, "get_entity", side_effect=Exception("Solr down")):
        assert SolrImplementation()._get_entity_name("MONDO:0007078") == "MONDO:0007078"



# =====================================================================
# Tests for get_case_phenotype_matrix
# =====================================================================


def test_case_phenotype_matrix_empty():
    with (
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Disease"),
    ):
        mock_query.return_value = {"response": {"docs": []}}
        result = SolrImplementation().get_case_phenotype_matrix("MONDO:0007078")
        assert isinstance(result, CasePhenotypeMatrixResponse)
        assert result.total_cases == 0


def test_case_phenotype_matrix_limit_exceeded():
    with patch.object(SolrImplementation, "_raw_solr_query") as mock_query:
        mock_query.return_value = {
            "response": {"docs": [{"subject": f"CASE:{i}", "object": "MONDO:0007078"} for i in range(11)]}
        }
        with pytest.raises(ValueError, match="exceeds limit"):
            SolrImplementation().get_case_phenotype_matrix("MONDO:0007078", limit=10)


def test_case_phenotype_matrix_complete_flow():
    responses = [
        {
            "response": {
                "docs": [
                    {"subject": "CASE:001", "subject_label": "Case 1", "object": "MONDO:0007078", "object_label": "D"}
                ]
            }
        },
        {
            "response": {
                "docs": [
                    {
                        "subject": "CASE:001",
                        "object": "HP:001",
                        "object_label": "P1",
                        "object_closure": ["HP:001", "UPHENO:0001001"],
                    }
                ]
            },
            "facet_counts": {"facet_queries": {}},
        },
    ]
    call_idx = [0]

    def side_effect(params):
        i = call_idx[0]
        call_idx[0] += 1
        return responses[i]

    with (
        patch.object(SolrImplementation, "_raw_solr_query", side_effect=side_effect),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Disease"),
    ):
        result = SolrImplementation().get_case_phenotype_matrix("MONDO:0007078")
        assert result.total_cases == 1
        assert result.disease_id == "MONDO:0007078"


# =====================================================================
# Tests for get_entity_grid
# =====================================================================


def test_entity_grid_empty():
    with (
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Entity"),
    ):
        mock_query.return_value = {"response": {"docs": []}}
        result = SolrImplementation().get_entity_grid(context_id="MONDO:0007078", grid_type="case-phenotype")
        assert isinstance(result, EntityGridResponse)
        assert result.total_columns == 0


def test_entity_grid_limit_exceeded():
    with patch.object(SolrImplementation, "_raw_solr_query") as mock_query:
        mock_query.return_value = {
            "response": {"docs": [{"subject": f"CASE:{i}", "object": "MONDO:0007078"} for i in range(11)]}
        }
        with pytest.raises(ValueError, match="exceeds limit"):
            SolrImplementation().get_entity_grid(context_id="MONDO:0007078", grid_type="case-phenotype", limit=10)


def test_entity_grid_unknown_type():
    with pytest.raises(ValueError, match="Unknown grid type"):
        SolrImplementation().get_entity_grid(context_id="MONDO:0007078", grid_type="nonexistent-type")


def test_entity_grid_complete_flow():
    responses = [
        {"response": {"docs": [{"subject": "CASE:001", "subject_label": "Case 1", "object": "MONDO:0007078"}]}},
        {
            "response": {
                "docs": [
                    {
                        "subject": "CASE:001",
                        "object": "HP:001",
                        "object_label": "P1",
                        "object_closure": ["HP:001", "UPHENO:0001001"],
                    }
                ]
            },
            "facet_counts": {"facet_queries": {}},
        },
    ]
    call_idx = [0]

    def side_effect(params):
        i = call_idx[0]
        call_idx[0] += 1
        return responses[i]

    with (
        patch.object(SolrImplementation, "_raw_solr_query", side_effect=side_effect),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Entity"),
    ):
        result = SolrImplementation().get_entity_grid(context_id="MONDO:0007078", grid_type="case-phenotype")
        assert result.total_columns == 1
        assert result.context_id == "MONDO:0007078"


# =====================================================================
# Helper for generic entity grid tests
# =====================================================================


def _mock_gene_entity(name="Test Gene"):
    mock = MagicMock()
    mock.category = "biolink:Gene"
    mock.configure_mock(name=name)
    return mock


# =====================================================================
# Tests for get_generic_entity_grid
# =====================================================================


def test_generic_grid_empty():
    with (
        patch.object(SolrImplementation, "get_entity", return_value=_mock_gene_entity()),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Gene"),
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
    ):
        mock_query.return_value = {"response": {"docs": []}}
        result = SolrImplementation().get_generic_entity_grid(
            context_id="HGNC:4851",
            column_assoc_categories=["biolink:CausalGeneToDiseaseAssociation"],
            row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
        )
        assert result.total_columns == 0


def test_generic_grid_limit_exceeded():
    with (
        patch.object(SolrImplementation, "get_entity", return_value=_mock_gene_entity()),
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
    ):
        mock_query.return_value = {
            "response": {"docs": [{"object": f"MONDO:{i:07d}", "subject": "HGNC:4851"} for i in range(11)]}
        }
        with pytest.raises(ValueError, match="exceeds limit"):
            SolrImplementation().get_generic_entity_grid(
                context_id="HGNC:4851",
                column_assoc_categories=["biolink:CausalGeneToDiseaseAssociation"],
                row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
                limit=10,
            )


def test_generic_grid_context_not_found():
    with patch.object(SolrImplementation, "get_entity", return_value=None):
        with pytest.raises(ValueError, match="not found"):
            SolrImplementation().get_generic_entity_grid(
                context_id="HGNC:NONEXISTENT",
                column_assoc_categories=["biolink:CausalGeneToDiseaseAssociation"],
                row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
            )


def test_generic_grid_field_mapping():
    with (
        patch.object(SolrImplementation, "get_entity", return_value=_mock_gene_entity()),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Gene"),
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
    ):
        mock_query.return_value = {"response": {"docs": []}}
        result = SolrImplementation().get_generic_entity_grid(
            context_id="HGNC:4851",
            column_assoc_categories=["biolink:CausalGeneToDiseaseAssociation"],
            row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
        )
        assert result is not None


def test_generic_grid_column_sorting():
    col_result = {
        "response": {
            "docs": [
                {
                    "object": "MONDO:001",
                    "object_label": "D1",
                    "subject": "HGNC:4851",
                    "subject_label": "HTT",
                    "category": "biolink:CorrelatedGeneToDiseaseAssociation",
                },
                {
                    "object": "MONDO:002",
                    "object_label": "D2",
                    "subject": "HGNC:4851",
                    "subject_label": "HTT",
                    "category": "biolink:CausalGeneToDiseaseAssociation",
                },
            ]
        }
    }
    row_result = {"response": {"docs": []}, "facet_counts": {"facet_queries": {}}}
    call_idx = [0]

    def side_effect(params):
        i = call_idx[0]
        call_idx[0] += 1
        return col_result if i == 0 else row_result

    with (
        patch.object(SolrImplementation, "get_entity", return_value=_mock_gene_entity("HTT")),
        patch.object(SolrImplementation, "_get_entity_name", return_value="HTT"),
        patch.object(SolrImplementation, "_raw_solr_query", side_effect=side_effect),
    ):
        result = SolrImplementation().get_generic_entity_grid(
            context_id="HGNC:4851",
            column_assoc_categories=[
                "biolink:CausalGeneToDiseaseAssociation",
                "biolink:CorrelatedGeneToDiseaseAssociation",
            ],
            row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
            group_columns_by_category=True,
        )
        assert result.columns[0].source_association_category == "biolink:CausalGeneToDiseaseAssociation"
        assert result.columns[1].source_association_category == "biolink:CorrelatedGeneToDiseaseAssociation"


@pytest.mark.parametrize("row_grouping", ["histopheno", "none"])
def test_generic_grid_row_grouping(row_grouping):
    with (
        patch.object(SolrImplementation, "get_entity", return_value=_mock_gene_entity()),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Gene"),
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
    ):
        mock_query.return_value = {"response": {"docs": []}}
        result = SolrImplementation().get_generic_entity_grid(
            context_id="HGNC:4851",
            column_assoc_categories=["biolink:CausalGeneToDiseaseAssociation"],
            row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
            row_grouping=row_grouping,
        )
        assert result is not None


def test_generic_grid_entity_list_category():
    mock_entity = _mock_gene_entity()
    mock_entity.category = ["biolink:Gene", "biolink:NamedThing"]

    with (
        patch.object(SolrImplementation, "get_entity", return_value=mock_entity),
        patch.object(SolrImplementation, "_get_entity_name", return_value="Test Gene"),
        patch.object(SolrImplementation, "_raw_solr_query") as mock_query,
    ):
        mock_query.return_value = {"response": {"docs": []}}
        result = SolrImplementation().get_generic_entity_grid(
            context_id="HGNC:4851",
            column_assoc_categories=["biolink:CausalGeneToDiseaseAssociation"],
            row_assoc_categories=["biolink:DiseaseToPhenotypicFeatureAssociation"],
        )
        assert result is not None



"""Tests for HistoPheno bin label mapping."""
import pytest
from monarch_py.datamodels.solr import (
    HistoPhenoKeys,
    HISTOPHENO_BIN_LABELS,
    get_bin_label,
    get_all_bin_ids,
)


class TestHistoPhenoBinLabels:
    """Test that all HistoPhenoKeys have human-readable labels."""

    def test_all_keys_have_labels(self):
        """Every HistoPhenoKeys enum should have a corresponding label."""
        for key in HistoPhenoKeys:
            assert key.value in HISTOPHENO_BIN_LABELS, f"Missing label for {key}"

    def test_labels_are_nonempty_strings(self):
        """Labels should be meaningful strings."""
        for key_value, label in HISTOPHENO_BIN_LABELS.items():
            assert isinstance(label, str)
            assert len(label) > 0, f"Empty label for {key_value}"

    def test_label_count_matches_enum_count(self):
        """Number of labels should match number of enum values."""
        assert len(HISTOPHENO_BIN_LABELS) == len(HistoPhenoKeys)

    @pytest.mark.parametrize(
        "bin_id,expected_label",
        [
            ("UPHENO:0002964", "skeletal system"),
            ("UPHENO:0004523", "nervous system"),
            ("HP:0001939", "metabolism/homeostasis"),
            ("UPHENO:0080362", "cardiovascular system"),
            ("UPHENO:0004536", "respiratory system"),
            ("UPHENO:0002833", "digestive system"),
            ("UPHENO:0002635", "integument"),
            ("UPHENO:0003020", "eye"),
            ("HP:0000598", "ear"),
            ("UPHENO:0003116", "endocrine system"),
            ("UPHENO:0002642", "genitourinary system"),
            ("UPHENO:0002816", "musculature"),
            ("UPHENO:0002764", "head/neck"),
            ("UPHENO:0002948", "immune system"),
            ("UPHENO:0004459", "blood/blood-forming tissues"),
            ("HP:0002664", "neoplasm"),
            ("UPHENO:0075949", "prenatal/birth"),
            ("UPHENO:0049874", "growth"),
            ("UPHENO:0003013", "breast"),
            ("UPHENO:0002712", "connective tissue"),
        ],
    )
    def test_known_labels(self, bin_id, expected_label):
        """Verify expected labels for HistoPheno bins."""
        assert HISTOPHENO_BIN_LABELS[bin_id] == expected_label


class TestGetBinLabel:
    """Test the get_bin_label helper function."""

    def test_get_valid_bin_label(self):
        """get_bin_label returns correct label for valid bin ID."""
        assert get_bin_label("UPHENO:0002964") == "skeletal system"

    def test_get_bin_label_raises_for_invalid_id(self):
        """get_bin_label raises KeyError for unknown bin ID."""
        with pytest.raises(KeyError):
            get_bin_label("INVALID:0000000")


class TestGetAllBinIds:
    """Test the get_all_bin_ids helper function."""

    def test_returns_all_bin_ids(self):
        """get_all_bin_ids returns all bin IDs."""
        bin_ids = get_all_bin_ids()
        assert len(bin_ids) == len(HistoPhenoKeys)

    def test_returns_list(self):
        """get_all_bin_ids returns a list."""
        bin_ids = get_all_bin_ids()
        assert isinstance(bin_ids, list)

    def test_all_ids_are_strings(self):
        """All returned IDs should be strings."""
        bin_ids = get_all_bin_ids()
        for bin_id in bin_ids:
            assert isinstance(bin_id, str)

    def test_all_ids_are_valid_enum_values(self):
        """All returned IDs should be valid HistoPhenoKeys values."""
        bin_ids = get_all_bin_ids()
        enum_values = {key.value for key in HistoPhenoKeys}
        for bin_id in bin_ids:
            assert bin_id in enum_values, f"{bin_id} not in HistoPhenoKeys"



"""API endpoints for entity grids.

Provides generic entity x entity grid endpoints that support:
- Disease -> Case × Phenotype (case-phenotype-grid)
- Gene -> Disease × Phenotype (disease-phenotype-grid)
- Gene -> Ortholog × Phenotype (ortholog-phenotype-grid)
- Generic grids with configurable association categories
"""
import logging
from enum import Enum
from typing import List
from fastapi import APIRouter, HTTPException, Query, Path

from monarch_py.api.config import solr
from monarch_py.datamodels.model import EntityGridResponse
from monarch_py.datamodels.category_enums import AssociationCategory, EntityCategory
from monarch_py.utils.association_type_utils import AssociationTypeMappings

logger = logging.getLogger(__name__)


# Define row grouping options
class RowGrouping(str, Enum):
    """Row grouping options."""
    HISTOPHENO = "histopheno"
    NONE = "none"

router = APIRouter(
    tags=["entity-grid"],
    responses={
        400: {"description": "Column count exceeds limit, or ID category mismatch"},
        404: {"description": "Entity not found"},
        422: {"description": "Invalid entity ID format"},
    },
)


def _validate_entity_category(entity_id: str, expected_category: EntityCategory) -> None:
    """Validate that an entity exists and has the expected category.

    Args:
        entity_id: The entity ID to validate
        expected_category: The expected biolink category

    Raises:
        HTTPException: If entity not found or wrong category
    """
    solr_impl = solr()
    try:
        entity = solr_impl.get_entity(entity_id, extra=False)
        if entity is None:
            raise HTTPException(
                status_code=404,
                detail=f"Entity {entity_id} not found",
            )

        # Check category
        category = getattr(entity, "category", None)
        if category is None:
            raise HTTPException(
                status_code=400,
                detail=f"Entity {entity_id} has no category",
            )

        # Category can be a string or list
        categories = [category] if isinstance(category, str) else category
        if expected_category.value not in categories:
            raise HTTPException(
                status_code=400,
                detail=f"{entity_id} is not a {expected_category.name.lower().replace('_', ' ')}. "
                f"Expected category: {expected_category.value}",
            )
    except HTTPException:
        raise
    except Exception as exc:
        logger.error("Error validating entity %s: %s", entity_id, exc)
        raise HTTPException(
            status_code=404,
            detail=f"Entity {entity_id} not found",
        )


def _get_grid(
    context_id: str,
    grid_type: str,
    direct: bool,
    limit: int,
) -> EntityGridResponse:
    """Get entity grid using the generic implementation.

    Args:
        context_id: The context entity ID
        grid_type: The grid type identifier
        direct: If True, only direct associations
        limit: Maximum column entities

    Returns:
        EntityGridResponse

    Raises:
        HTTPException: On errors
    """
    solr_impl = solr()
    try:
        return solr_impl.get_entity_grid(
            context_id=context_id,
            grid_type=grid_type,
            direct_only=direct,
            limit=limit,
        )
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))


@router.get(
    "/{context_id}/case-phenotype-grid",
    response_model=EntityGridResponse,
    summary="Get case-phenotype grid for a disease",
    description="""
    Returns a grid of cases (patients) and their phenotypes for a given disease.

    The grid includes:
    - **columns**: List of case entities with the disease
    - **rows**: List of phenotypes observed across all cases
    - **bins**: Body system categories (HistoPheno) with phenotype counts
    - **cells**: Map of case:phenotype pairs to observation details

    Use `direct=true` (default) to get only cases with exactly this disease.
    Use `direct=false` to include cases with descendant diseases (subtypes).
    """,
)
async def get_case_phenotype_grid(
    context_id: str = Path(
        ...,
        description="MONDO disease ID",
        examples=["MONDO:0007078"],
    ),
    direct: bool = Query(
        True,
        description="Only include cases with exactly this disease",
    ),
    limit: int = Query(
        1000,
        ge=1,
        le=1000,
        description="Maximum number of cases (columns)",
    ),
) -> EntityGridResponse:
    """Get case-phenotype grid for a disease.

    The context_id must be a disease ID (e.g., MONDO:0007078).
    """
    _validate_entity_category(context_id, EntityCategory.DISEASE)
    return _get_grid(context_id, "case-phenotype", direct, limit)


@router.get(
    "/{context_id}/disease-phenotype-grid",
    response_model=EntityGridResponse,
    summary="Get disease-phenotype grid for a gene",
    description="""
    Returns a grid of diseases and their phenotypes for a given gene.

    The grid shows diseases causally associated with the gene and the
    phenotypes associated with each disease.

    - **columns**: List of disease entities associated with the gene
    - **rows**: List of phenotypes observed across all diseases
    - **bins**: Body system categories (HistoPheno) with phenotype counts
    - **cells**: Map of disease:phenotype pairs to observation details
    """,
)
async def get_disease_phenotype_grid(
    context_id: str = Path(
        ...,
        description="Gene ID (e.g., HGNC ID)",
        examples=["HGNC:4851"],
    ),
    direct: bool = Query(
        True,
        description="Only include directly associated diseases",
    ),
    limit: int = Query(
        500,
        ge=1,
        le=500,
        description="Maximum number of diseases (columns)",
    ),
) -> EntityGridResponse:
    """Get disease-phenotype grid for a gene.

    The context_id must be a gene ID (e.g., HGNC:4851).
    """
    _validate_entity_category(context_id, EntityCategory.GENE)
    return _get_grid(context_id, "disease-phenotype", direct, limit)


@router.get(
    "/{context_id}/ortholog-phenotype-grid",
    response_model=EntityGridResponse,
    summary="Get ortholog-phenotype grid for a gene",
    description="""
    Returns a grid of orthologs and their phenotypes for a given gene.

    The grid shows orthologous genes and the phenotypes associated with
    each ortholog (typically from model organisms).

    - **columns**: List of orthologous genes
    - **rows**: List of phenotypes observed across all orthologs
    - **bins**: Body system categories (HistoPheno) with phenotype counts
    - **cells**: Map of ortholog:phenotype pairs to observation details
    """,
)
async def get_ortholog_phenotype_grid(
    context_id: str = Path(
        ...,
        description="Gene ID (e.g., HGNC ID)",
        examples=["HGNC:4851"],
    ),
    direct: bool = Query(
        True,
        description="Only include direct orthologs",
    ),
    limit: int = Query(
        500,
        ge=1,
        le=500,
        description="Maximum number of orthologs (columns)",
    ),
) -> EntityGridResponse:
    """Get ortholog-phenotype grid for a gene.

    The context_id must be a gene ID (e.g., HGNC:4851).
    """
    _validate_entity_category(context_id, EntityCategory.GENE)
    return _get_grid(context_id, "ortholog-phenotype", direct, limit)


# =============================================================================
# Generic Entity Grid Endpoint
# =============================================================================


@router.get(
    "-grid/{context_id}",
    response_model=EntityGridResponse,
    summary="Get generic entity grid with configurable associations",
    description="""
    Returns a generic entity × entity grid with configurable association categories.

    This endpoint allows you to specify:
    - **column_association_category**: One or more association categories for the
      context → column hop (e.g., causal gene-disease, correlated gene-disease)
    - **row_association_category**: The association category for the column → row hop
    - **row_grouping**: How to group rows (histopheno for body systems, or none)
    - **group_columns_by_category**: Whether to sort columns by their source association

    Example: To get a grid of diseases (causal + correlated) and their phenotypes for a gene:
    ```
    GET /entity-grid/HGNC:4851?column_association_category=biolink:CausalGeneToDiseaseAssociation
        &column_association_category=biolink:CorrelatedGeneToDiseaseAssociation
        &row_association_category=biolink:DiseaseToPhenotypicFeatureAssociation
        &group_columns_by_category=true
    ```

    The response includes source_association_category on each column entity,
    allowing the frontend to display different column groups.
    """,
)
async def get_generic_entity_grid(
    context_id: str = Path(
        ...,
        description="Context entity ID (gene, disease, etc.)",
        examples=["HGNC:4851", "MONDO:0007078"],
    ),
    column_association_category: List[AssociationCategory] = Query(
        ...,
        description="Association category/categories for context → column",
        examples=["biolink:CausalGeneToDiseaseAssociation"],
    ),
    row_association_category: List[AssociationCategory] = Query(
        ...,
        description="Association category/categories for column → row",
        examples=["biolink:DiseaseToPhenotypicFeatureAssociation"],
    ),
    row_grouping: RowGrouping = Query(
        RowGrouping.HISTOPHENO,
        description="How to group rows",
    ),
    group_columns_by_category: bool = Query(
        False,
        description="Sort columns by association category",
    ),
    direct: bool = Query(
        True,
        description="Only include direct associations (not via closure)",
    ),
    limit: int = Query(
        500,
        ge=1,
        le=1000,
        description="Maximum number of column entities",
    ),
) -> EntityGridResponse:
    """Get a generic entity grid with configurable association categories.

    This endpoint dynamically constructs a grid based on the specified
    association categories, allowing flexible exploration of relationships.
    """
    solr_impl = solr()

    # Convert enum values to strings
    column_categories = [cat.value for cat in column_association_category]
    row_categories = [cat.value for cat in row_association_category]

    try:
        return solr_impl.get_generic_entity_grid(
            context_id=context_id,
            column_assoc_categories=column_categories,
            row_assoc_categories=row_categories,
            row_grouping=row_grouping.value,
            group_columns_by_category=group_columns_by_category,
            direct_only=direct,
            limit=limit,
        )
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))


# =============================================================================
# Traversable Associations Metadata Endpoint
# =============================================================================


@router.get(
    "-grid/traversable-associations/{entity_category}",
    summary="Get traversable associations for an entity type",
    description="""
    Returns association types that can be traversed from a given entity category.

    For each association, returns:
    - **category**: The biolink association category
    - **label**: Human-readable label for UI display
    - **context_field**: Which field ("subject" or "object") the context entity occupies
    - **target_category**: The biolink category of entities at the other end

    This enables dynamic UI that shows only valid association options based on
    the selected entity type, supporting bidirectional traversal of associations.

    Example: For biolink:Gene, this returns gene-disease, gene-phenotype, gene-homology, etc.
    For biolink:Disease, this returns disease-phenotype, and also gene-disease (reverse direction).
    """,
)
async def get_traversable_associations(
    entity_category: str = Path(
        ...,
        description="The biolink category of the context entity",
        examples=["biolink:Gene", "biolink:Disease", "biolink:Case"],
    ),
) -> List[dict]:
    """Get associations traversable from a given entity category.

    Returns a list of association options with direction information,
    enabling the frontend to build dynamic association selection UIs.
    """
    associations = AssociationTypeMappings.get_traversable_associations(entity_category)
    if not associations:
        raise HTTPException(
            status_code=404,
            detail=f"No traversable associations found for entity category: {entity_category}",
        )
    return associations



"""Unit tests for entity grid utilities."""

from monarch_py.utils.entity_grid_utils import (
    _build_columns,
    _build_rows,
    _build_cells,
    _build_bins,
    _find_bin_for_entity,
    build_entity_grid,
    make_cell_key,
    sort_columns_by_category,
)
from monarch_py.datamodels.grid_configs import get_grid_config
from monarch_py.datamodels.grid_groupings import RowGroupingConfig, GroupingType
from monarch_py.datamodels.model import GridColumnEntity, EntityGridResponse


# =====================================================================
# Tests for _build_columns
# =====================================================================


def test_columns_include_source_association_data():
    config = get_grid_config("case-phenotype")
    docs = [
        {
            "subject": "CASE:001",
            "subject_label": "Case 1",
            "object": "MONDO:0007078",
            "object_label": "Achondroplasia",
            "category": "biolink:CaseToDiseaseAssociation",
            "predicate": "biolink:has_phenotype",
            "publications": ["PMID:12345"],
            "primary_knowledge_source": "infores:hpo",
        },
        {
            "subject": "CASE:002",
            "subject_label": "Case 2",
            "object": "MONDO:0007078",
            "object_label": "Achondroplasia",
            "category": "biolink:CaseToDiseaseAssociation",
            "predicate": "biolink:has_phenotype",
            "publications": ["PMID:67890", "PMID:11111"],
            "primary_knowledge_source": "infores:hpo",
        },
    ]

    columns = _build_columns(docs, "MONDO:0007078", config)

    assert len(columns) == 2
    col1 = columns[0]
    assert col1.source_association_category == "biolink:CaseToDiseaseAssociation"
    assert col1.source_association_predicate == "biolink:has_phenotype"
    assert col1.source_association_publications == ["PMID:12345"]
    assert col1.source_association_primary_knowledge_source == "infores:hpo"


def test_columns_handle_missing_association_fields():
    config = get_grid_config("case-phenotype")
    docs = [
        {
            "subject": "CASE:001",
            "object": "MONDO:0007078",
        }
    ]

    columns = _build_columns(docs, "MONDO:0007078", config)

    assert len(columns) == 1
    assert columns[0].source_association_category is None
    assert columns[0].source_association_predicate is None


# =====================================================================
# Tests for sort_columns_by_category
# =====================================================================


def test_sort_columns_by_category():
    columns = [
        GridColumnEntity(
            id="MONDO:001",
            category="biolink:Disease",
            is_direct=True,
            source_association_category="biolink:CorrelatedGeneToDiseaseAssociation",
        ),
        GridColumnEntity(
            id="MONDO:002",
            category="biolink:Disease",
            is_direct=True,
            source_association_category="biolink:CausalGeneToDiseaseAssociation",
        ),
        GridColumnEntity(
            id="MONDO:003",
            category="biolink:Disease",
            is_direct=True,
            source_association_category="biolink:CorrelatedGeneToDiseaseAssociation",
        ),
    ]

    category_order = [
        "biolink:CausalGeneToDiseaseAssociation",
        "biolink:CorrelatedGeneToDiseaseAssociation",
    ]

    sorted_cols = sort_columns_by_category(columns, category_order)
    assert sorted_cols[0].id == "MONDO:002"
    assert sorted_cols[1].source_association_category == "biolink:CorrelatedGeneToDiseaseAssociation"
    assert sorted_cols[2].source_association_category == "biolink:CorrelatedGeneToDiseaseAssociation"


def test_sort_preserves_order_within_category():
    columns = [
        GridColumnEntity(
            id="A", category="biolink:Disease", is_direct=True, source_association_category="biolink:Causal"
        ),
        GridColumnEntity(
            id="B", category="biolink:Disease", is_direct=True, source_association_category="biolink:Causal"
        ),
        GridColumnEntity(
            id="C", category="biolink:Disease", is_direct=True, source_association_category="biolink:Causal"
        ),
    ]

    sorted_cols = sort_columns_by_category(columns, ["biolink:Causal"])
    assert [c.id for c in sorted_cols] == ["A", "B", "C"]


def test_sort_empty_list():
    assert sort_columns_by_category([], ["biolink:Causal"]) == []


def test_sort_empty_category_order():
    columns = [
        GridColumnEntity(id="A", category="biolink:Disease", is_direct=True),
        GridColumnEntity(id="B", category="biolink:Disease", is_direct=True),
    ]
    sorted_cols = sort_columns_by_category(columns, [])
    assert [c.id for c in sorted_cols] == ["A", "B"]


# =====================================================================
# Tests for make_cell_key
# =====================================================================


def test_make_cell_key():
    assert make_cell_key("COL:001", "ROW:001") == "COL:001:ROW:001"


def test_make_cell_key_with_different_ids():
    assert make_cell_key("CASE:123", "HP:0001234") == "CASE:123:HP:0001234"


# =====================================================================
# Tests for _find_bin_for_entity
# =====================================================================


def test_find_bin_returns_matching():
    assert (
        _find_bin_for_entity(["HP:root", "BIN:001", "HP:child"], {"BIN:001", "BIN:002"}, ["BIN:001", "BIN:002"])
        == "BIN:001"
    )


def test_find_bin_returns_none_when_no_match():
    assert _find_bin_for_entity(["HP:root", "HP:child"], {"BIN:001", "BIN:002"}, ["BIN:001", "BIN:002"]) is None


def test_find_bin_returns_first_ordered_when_multiple():
    assert _find_bin_for_entity(["BIN:002", "BIN:001"], {"BIN:001", "BIN:002"}, ["BIN:002", "BIN:001"]) == "BIN:002"


def test_find_bin_empty_closure():
    assert _find_bin_for_entity([], {"BIN:001"}, ["BIN:001"]) is None


def test_find_bin_empty_bin_ids():
    assert _find_bin_for_entity(["HP:root", "BIN:001"], set(), []) is None


# =====================================================================
# Tests for _build_rows
# =====================================================================


def test_build_rows_from_docs():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS,
        bin_ids=["BIN:001", "BIN:002"],
        bin_labels={"BIN:001": "Bin 1", "BIN:002": "Bin 2"},
    )
    docs = [
        {
            "object": "HP:001",
            "object_label": "Phenotype 1",
            "object_closure": ["HP:root", "BIN:001", "HP:001"],
            "subject": "CASE:001",
        },
        {
            "object": "HP:002",
            "object_label": "Phenotype 2",
            "object_closure": ["HP:root", "BIN:002", "HP:002"],
            "subject": "CASE:001",
        },
    ]
    rows = _build_rows(docs, config, grouping)
    assert len(rows) == 2
    assert rows[0].id == "HP:001"
    assert rows[0].bin_id == "BIN:001"
    assert rows[1].id == "HP:002"
    assert rows[1].bin_id == "BIN:002"


def test_build_rows_deduplicates():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={"BIN:001": "Bin 1"}
    )
    docs = [
        {"object": "HP:001", "object_label": "P1", "object_closure": ["BIN:001"], "subject": "CASE:001"},
        {"object": "HP:001", "object_label": "P1", "object_closure": ["BIN:001"], "subject": "CASE:002"},
    ]
    rows = _build_rows(docs, config, grouping)
    assert len(rows) == 1


def test_build_rows_skips_no_bin_match():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={"BIN:001": "Bin 1"}
    )
    docs = [{"object": "HP:001", "object_label": "P1", "object_closure": ["HP:root"], "subject": "CASE:001"}]
    rows = _build_rows(docs, config, grouping)
    assert len(rows) == 0


def test_build_rows_skips_missing_row_id():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={"BIN:001": "Bin 1"}
    )
    docs = [{"subject": "CASE:001", "object_closure": ["BIN:001"]}]
    rows = _build_rows(docs, config, grouping)
    assert len(rows) == 0


# =====================================================================
# Tests for _build_cells
# =====================================================================


def test_build_cells_from_docs():
    config = get_grid_config("case-phenotype")
    column_map = {
        "CASE:001": GridColumnEntity(id="CASE:001", category="biolink:Case", is_direct=True),
        "CASE:002": GridColumnEntity(id="CASE:002", category="biolink:Case", is_direct=True),
    }
    docs = [
        {"subject": "CASE:001", "object": "HP:001"},
        {"subject": "CASE:002", "object": "HP:001"},
    ]
    cells = _build_cells(docs, column_map, config)
    assert len(cells) == 2
    assert "CASE:001:HP:001" in cells
    assert cells["CASE:001:HP:001"].present is True


def test_build_cells_with_qualifiers():
    config = get_grid_config("case-phenotype")
    column_map = {"CASE:001": GridColumnEntity(id="CASE:001", category="biolink:Case", is_direct=True)}
    docs = [
        {
            "subject": "CASE:001",
            "object": "HP:001",
            "onset_qualifier": "HP:0003577",
            "onset_qualifier_label": "Congenital onset",
        }
    ]
    cells = _build_cells(docs, column_map, config)
    cell = cells["CASE:001:HP:001"]
    assert cell.qualifiers is not None
    assert cell.qualifiers["onset_qualifier"] == "HP:0003577"


def test_build_cells_without_qualifiers():
    config = get_grid_config("case-phenotype")
    column_map = {"CASE:001": GridColumnEntity(id="CASE:001", category="biolink:Case", is_direct=True)}
    docs = [{"subject": "CASE:001", "object": "HP:001"}]
    cells = _build_cells(docs, column_map, config)
    assert cells["CASE:001:HP:001"].qualifiers is None


def test_build_cells_negated():
    config = get_grid_config("case-phenotype")
    column_map = {"CASE:001": GridColumnEntity(id="CASE:001", category="biolink:Case", is_direct=True)}
    docs = [{"subject": "CASE:001", "object": "HP:001", "negated": True}]
    cells = _build_cells(docs, column_map, config)
    assert cells["CASE:001:HP:001"].negated is True


def test_build_cells_skips_missing_ids():
    config = get_grid_config("case-phenotype")
    column_map = {"CASE:001": GridColumnEntity(id="CASE:001", category="biolink:Case", is_direct=True)}
    docs = [{"subject": "CASE:001"}, {"object": "HP:001"}]
    cells = _build_cells(docs, column_map, config)
    assert len(cells) == 0


def test_build_cells_skips_unknown_column():
    config = get_grid_config("case-phenotype")
    column_map = {"CASE:001": GridColumnEntity(id="CASE:001", category="biolink:Case", is_direct=True)}
    docs = [{"subject": "CASE:999", "object": "HP:001"}]
    cells = _build_cells(docs, column_map, config)
    assert len(cells) == 0


# =====================================================================
# Tests for _build_bins
# =====================================================================


def test_build_bins_from_facet_counts():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS,
        bin_ids=["BIN:001", "BIN:002"],
        bin_labels={"BIN:001": "Bin One", "BIN:002": "Bin Two"},
    )
    facet_counts = {'object_closure:"BIN:001"': 5, 'object_closure:"BIN:002"': 3}
    bins = _build_bins(facet_counts, grouping, config)
    assert len(bins) == 2
    assert bins[0].id == "BIN:001"
    assert bins[0].label == "Bin One"
    assert bins[0].count == 5
    assert bins[1].count == 3


def test_build_bins_zero_counts():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={"BIN:001": "Bin 1"}
    )
    bins = _build_bins({}, grouping, config)
    assert len(bins) == 1
    assert bins[0].count == 0


def test_build_bins_id_as_label_fallback():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={})
    bins = _build_bins({'object_closure:"BIN:001"': 1}, grouping, config)
    assert bins[0].label == "BIN:001"


# =====================================================================
# Tests for build_entity_grid
# =====================================================================


def test_build_entity_grid_complete():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={"BIN:001": "Bin 1"}
    )

    column_docs = [
        {"subject": "CASE:001", "subject_label": "Case 1", "object": "MONDO:0007078", "object_label": "Achondroplasia"}
    ]
    row_docs = [
        {
            "subject": "CASE:001",
            "object": "HP:001",
            "object_label": "Phenotype 1",
            "object_closure": ["BIN:001", "HP:001"],
        }
    ]
    facet_counts = {'object_closure:"BIN:001"': 1}

    grid = build_entity_grid(
        context_id="MONDO:0007078",
        context_name="Achondroplasia",
        context_category="biolink:Disease",
        config=config,
        grouping=grouping,
        column_docs=column_docs,
        row_docs=row_docs,
        facet_counts=facet_counts,
    )

    assert isinstance(grid, EntityGridResponse)
    assert grid.context_id == "MONDO:0007078"
    assert grid.total_columns == 1
    assert grid.total_rows == 1
    assert len(grid.bins) == 1
    assert "CASE:001:HP:001" in grid.cells


def test_build_entity_grid_empty():
    config = get_grid_config("case-phenotype")
    grouping = RowGroupingConfig(
        grouping_type=GroupingType.CLOSURE_ROOTS, bin_ids=["BIN:001"], bin_labels={"BIN:001": "Bin 1"}
    )

    grid = build_entity_grid(
        context_id="MONDO:0007078",
        context_name="Achondroplasia",
        context_category="biolink:Disease",
        config=config,
        grouping=grouping,
        column_docs=[],
        row_docs=[],
        facet_counts={},
    )

    assert grid.total_columns == 0
    assert grid.total_rows == 0
    assert len(grid.cells) == 0



"""Tests for case-phenotype matrix data models."""

import pytest
from monarch_py.datamodels.model import (
    CasePhenotypeMatrixResponse,
    CaseEntity,
    CasePhenotype,
    HistoPhenoBin,
    CasePhenotypeCellData,
)


class TestCaseEntity:
    """Test CaseEntity model."""

    def test_create_direct_case(self):
        """A direct case should have is_direct=True and no source_disease."""
        case = CaseEntity(
            id="MONARCH:case123",
            label="Patient 1",
            is_direct=True,
        )
        assert case.id == "MONARCH:case123"
        assert case.is_direct is True
        assert case.source_disease_id is None

    def test_create_indirect_case(self):
        """An indirect case should have is_direct=False and source disease info."""
        case = CaseEntity(
            id="MONARCH:case456",
            label="Patient 2",
            is_direct=False,
            source_disease_id="MONDO:0007079",
            source_disease_label="Specific Disease Subtype",
        )
        assert case.is_direct is False
        assert case.source_disease_id == "MONDO:0007079"


class TestCasePhenotype:
    """Test CasePhenotype model."""

    def test_phenotype_with_bin(self):
        """Phenotype should have bin assignment."""
        phenotype = CasePhenotype(
            id="HP:0001250",
            label="Seizure",
            bin_id="UPHENO:0004523",  # nervous system
        )
        assert phenotype.bin_id == "UPHENO:0004523"


class TestHistoPhenoBin:
    """Test HistoPhenoBin model."""

    def test_bin_with_count(self):
        """Bin should track phenotype count."""
        bin = HistoPhenoBin(
            id="UPHENO:0002964",
            label="skeletal system",
            phenotype_count=42,
        )
        assert bin.phenotype_count == 42


class TestCasePhenotypeCellData:
    """Test CasePhenotypeCellData model."""

    @pytest.mark.parametrize(
        "cell_kwargs,expected_attr,expected_value",
        [
            ({"present": True}, "present", True),
            ({"present": True}, "negated", None),
            ({"present": True, "negated": True}, "negated", True),
            ({"present": True, "negated": False}, "negated", False),
            ({"present": True, "onset_qualifier": "P2Y"}, "onset_qualifier", "P2Y"),
            ({"present": True, "onset_qualifier": "P6M"}, "onset_qualifier", "P6M"),
            ({"present": True, "onset_qualifier_label": "2 years"}, "onset_qualifier_label", "2 years"),
        ],
    )
    def test_cell_attributes(self, cell_kwargs, expected_attr, expected_value):
        """Cell should store various attributes correctly."""
        cell = CasePhenotypeCellData(**cell_kwargs)
        assert getattr(cell, expected_attr) == expected_value

    @pytest.mark.parametrize(
        "publications,expected_count",
        [
            (["PMID:12345"], 1),
            (["PMID:12345", "PMID:67890"], 2),
            ([], 0),
            (None, None),
        ],
    )
    def test_cell_publications(self, publications, expected_count):
        """Cell should handle publication lists."""
        cell = CasePhenotypeCellData(present=True, publications=publications)
        if expected_count is None:
            assert cell.publications is None
        else:
            assert len(cell.publications) == expected_count


class TestCasePhenotypeMatrixResponse:
    """Test complete matrix response model."""

    def test_empty_matrix(self):
        """Matrix with no cases or phenotypes."""
        matrix = CasePhenotypeMatrixResponse(
            disease_id="MONDO:0007078",
            disease_name="Test Disease",
            total_cases=0,
            total_phenotypes=0,
            cases=[],
            phenotypes=[],
            bins=[],
            cells={},
        )
        assert matrix.total_cases == 0
        assert len(matrix.cases) == 0

    def test_matrix_with_data(self):
        """Matrix with cases, phenotypes, and cells."""
        case = CaseEntity(id="MONARCH:case1", label="P1", is_direct=True)
        phenotype = CasePhenotype(id="HP:0001250", label="Seizure", bin_id="UPHENO:0004523")
        bin = HistoPhenoBin(id="UPHENO:0004523", label="nervous system", phenotype_count=1)
        cell_key = "MONARCH:case1:HP:0001250"
        cell = CasePhenotypeCellData(present=True)

        matrix = CasePhenotypeMatrixResponse(
            disease_id="MONDO:0007078",
            disease_name="Achondroplasia",
            total_cases=1,
            total_phenotypes=1,
            cases=[case],
            phenotypes=[phenotype],
            bins=[bin],
            cells={cell_key: cell},
        )
        assert matrix.total_cases == 1
        assert cell_key in matrix.cells
